{***************************************************************************}
{*                                                                         *}
{* Unit:VGAPal                                                 Version 3.0 *}
{*                    Created by Rafal Zielinski.April 1997                *}
{*                                                                         *}
{***************************************************************************}
{*                                                                         *}
{*  Features:                                                              *}
{*        -all basic and advanced palette routines !                       *}
{*        -allow change RGB in VGA 16 colors mode !                        *}
{*                                                                         *}
{***************************************************************************}
{$B-,F+,G+,I-,E-,P+,T-,V+,X+}

{$D-,L-,Q-,R-,S-,Y-}

Unit VGAPal;

Interface

Uses Crt;

Const bmDown   =1;
      bmMedium =2;
      bmUp     =3;
      bmFull   =4;

      efDark   =1;
      efLight  =2;
      efGrey   =3;
      efColor  =4;
      efBlue   =5;
      efGreen  =6;
      efRed    =7;
      efYellow =8;

     StandardPalette : array[0..15] of record
       R,G,B : Integer;
       Name: String[13];
       ColorNum: Integer;
     end = (
      ( R:0;  G:0;   B:0;   Name:'Black';	  ColorNum: 0),
      ( R:0;  G:0;   B:40;  Name:'Blue';	  ColorNum: 1),
      ( R:0;  G:40;  B:0;   Name:'Green';	  ColorNum: 2),
      ( R:0;  G:40;  B:40;  Name:'Cyan';	  ColorNum: 3),
      ( R:40; G:7;   B:7;   Name:'Red'; 	  ColorNum: 4),
      ( R:40; G:0;   B:40;  Name:'Magenta';	  ColorNum: 5),
      ( R:40; G:30;  B:0;   Name:'Brown';	  ColorNum: 20),
      ( R:49; G:49;  B:49;  Name:'LightGray';	  ColorNum: 7),
      ( R:26; G:26;  B:26;  Name:'Dark Gray';	  ColorNum: 56),
      ( R:0;  G:0;   B:63;  Name:'Light Blue';	  ColorNum: 57),
      ( R:9;  G:63;  B:9;   Name:'Light Green';   ColorNum: 58),
      ( R:0;  G:63;  B:63;  Name:'Light Cyan';	  ColorNum: 59),
      ( R:63; G:10;  B:10;  Name:'Light Red';	  ColorNum: 60),
      ( R:44; G:0;   B:63;  Name:'Light Magenta'; ColorNum: 61),
      ( R:63; G:63;  B:18;  Name:'Yellow';	  ColorNum: 62),
      ( R:63; G:63;  B:63;  Name: 'White';	  ColorNum: 63)
     );

Type TColor = array[0..2] of byte;

     TVGAPalette = array[0..255] of TColor;

     TLongColor = array[0..2] of integer;

     TLongVGAPalette = array[0..255] of TLongColor;

     TBlendInfo = record
		SPal: TVGAPalette;
		DPal: TVGAPalette;
		Time: word;
		Start: word;
		HowMany: word;
		Step: word;
     end;

     TEffectInfo = record
		 Effect: byte;
		 Save: Boolean;
		 Show: Boolean;
		 Time: word;
		 Start: word;
		 HowMany: word;
		 Step: word;
     end;

Var  BlackPal				:TVGAPalette;

Procedure SetRed(Color:TColor;AValue:byte);
Procedure SetGreen(Color:TColor;AValue:byte);
Procedure SetBlue(Color:TColor;AValue:byte);
Function GetRed(Color:TColor):byte;
Function GetGreen(Color:TColor):byte;
Function GetBlue(Color:TColor):byte;

Procedure SetCol(var Color:TColor;AR,AG,AB:byte);
Procedure SetColorRGB(nr:word;R,G,B:byte);
Procedure GetColorRGB(nr:word;Var R,G,B:byte);
Procedure SetAllPalette(var Pal:TVGAPalette);
Procedure SetPalettePortion(var Pal:TVGAPalette;Start,HowMany:word);
Procedure SetStandardPalette;
Procedure CopyStandardPalette(var Pal:TVGAPalette);
Procedure GetAllPalette(var Pal:TVGAPalette);
Procedure GetPalettePortion(var Pal:TVGAPalette;Start,HowMany:word);

Procedure SetRGBColorVGA16(nr:word;R,G,B:byte);

Function LoadPalette(name:string;var Pal:TVGAPalette):boolean;

Function NearColor(R,G,B:byte;Pal:TVGAPalette):byte;
Procedure ClearZeroColor(var Pal:TVGAPalette);

Procedure Blend(var BlendInfo:TBlendInfo;BlendMethod:byte);
Procedure BlendDown(var BlendInfo:TBlendInfo);
Procedure BlendMedium(var BlendInfo:TBlendInfo);
Procedure BlendUp(var BlendInfo:TBlendInfo);
Procedure BlendFull(var BlendInfo:TBlendInfo);

Procedure MakeEffect(var Pal:TVGAPalette;EffectInfo:TEffectInfo);

Implementation

Procedure SetRed;begin
	  Color[0]:=AVAlue;
	  end;

Procedure SetGreen;begin
	  Color[1]:=AVAlue;
	  end;

Procedure SetBlue;begin
	  Color[2]:=AVAlue;
	  end;

Function GetRed;begin
	 GetRed:=Color[0];
	 end;

Function GetGreen;begin
	 GetGreen:=Color[1];
	 end;

Function GetBlue;begin
	 GetBlue:=Color[2];
	 end;

Procedure SetCol;begin
	  Color[0]:=AR;
	  Color[1]:=AG;
	  Color[2]:=AB;
	  end;

Procedure SetColorRGB;assembler;
	  asm
	  mov ah,10h
	  mov al,10h
	  mov bx,[nr]
	  mov ch,[G]
	  mov cl,[B]
	  mov dh,[R]
	  int 10h
	  end;

Procedure GetColorRGB;assembler;
	  asm
	  mov ah,10h
	  mov al,15h
	  mov bx,[nr]
	  int 10h
	  les di,dword ptr [R]
	  mov [es:di],dh
	  les di,dword ptr [G]
	  mov [es:di],ch
	  les di,dword ptr [B]
	  mov [es:di],cl
	  end;

Procedure SetPalettePortion;assembler;
	  asm
	  push ds
	  lds si,[Pal]
	  mov ax,[Start]
	  shl ax,1
	  add ax,[Start]
	  add si,ax
	  mov dx,03C8h
	  mov ax,[Start]
	  out dx,al
	  inc dx
	  {cli}
	  push dx
	  mov cx,[HowMany]
	  shl cx,1
	  add cx,[HowMany]
	  mov dx,03DAh
@Loop1:   in al,dx
	  and al,08
	  jne @Loop1
@Loop2:   in al,dx
	  and al,08
	  je @Loop2
	  pop dx
	  rep outsb
	  {sti}
	  pop ds
	  end;

Procedure SetAllPalette;begin
	  SetPalettePortion(Pal,0,256);
	  end;

Procedure SetStandardPalette;
Var z:byte;
begin
	  For z:=0 to 15 do
	      SetColorRGB(z,StandardPalette[z].R,StandardPalette[z].G,StandardPalette[z].B);
	  end;

Procedure CopyStandardPalette;
Var z:byte;
begin
	  For z:=0 to 15 do begin
	      Pal[z,0]:=StandardPalette[z].R;
	      Pal[z,1]:=StandardPalette[z].G;
	      Pal[z,2]:=StandardPalette[z].B;
	  end;
	  end;

Procedure GetPalettePortion;assembler;
	  asm
	  les di,[Pal]
	  mov ax,[Start]
	  shl ax,1
	  add ax,[Start]
	  add si,ax
	  mov dx,03C7h
	  mov ax,[Start]
	  out dx,al
	  inc dx
	  inc dx
	  cli
	  push dx
	  mov cx,[HowMany]
	  shl cx,1
	  add cx,[HowMany]
	  mov dx,03DAh
@Loop1:   in al,dx
	  and al,08
	  jne @Loop1
@Loop2:   in al,dx
	  and al,08
	  je @Loop2
	  pop dx
	  rep insb
	  sti
	  end;

Procedure GetAllPalette;begin
	  GetPalettePortion(Pal,0,256);
	  end;

Procedure SetRGBColorVGA16(nr:word;R,G,B:byte);begin
	  SetColorRGB(StandardPalette[nr].ColorNum,R,G,B);
	  end;

Function LoadPalette;
Var f:file;
begin
	 LoadPalette:=true;
	 Assign(f,Name);
	 {$I-}
	 Reset(f,1);
	 {$I+}
	 If IOResult<>0 then begin
	    LoadPalette:=false;
	    exit;
	 end;
	 BlockRead(f,Pal,SizeOf(Pal));
	 Close(f);
	 end;

Function NearColor;
Var i:byte;
    Value,Diff,CurrentDiff:longint;
begin
     Value:=R*30+G*59+B*11;
     Diff:=65536;
     NearColor:=0;
     For i:=0 to 255 do begin
	 CurrentDiff:=Abs((Pal[i,0]*30+Pal[i,1]*59+Pal[i,2]*11)-Value);
	 if CurrentDiff<Diff then begin
	    NearColor:=i;Diff:=CurrentDiff
	 end;
     end;
     end;

Procedure ClearZeroColor;begin
	  SetCol(Pal[0],0,0,0);
	  end;

Procedure Blend;begin
	  Case BlendMethod of
	       bmDown:BlendDown(BlendInfo);
	       bmMedium:BlendMedium(BlendInfo);
	       bmUp:BlendUp(BlendInfo);
	       bmFull:BlendFull(BlendInfo);
	  end;
	  end;

Procedure Compare(var Source,Dest:byte);begin
	  if Source>Dest then Dec(Source) else
	     if Source<Dest then Inc(Source);
	  end;

Procedure BlendDown(var BlendInfo:TBlendInfo);
Var i,j,k:byte;
    Last:word;
begin
     With BlendInfo do begin
	  Last:=Start+HowMany-1;
	  For i:=0 to 63 do begin
	      For j:=Start to Last do
		  For k:=0 to 2 do
		      Compare(SPal[j,k],DPal[j,k]);
		      if (I MOD Step)=0 then
			 SetPalettePortion(SPAl,Start,HowMany);
		      Delay(Time);
	  end;
	  SetPalettePortion(DPal,Start,HowMany);
     end;
     end;

Procedure BlendUp(var BlendInfo:TBlendInfo);
Var i,j,k:byte;
    Last:word;
    HPal:TVGAPalette;
begin
     With BlendInfo do begin
	  Last:=Start+HowMAny-1;
	  For j:=Start to Last do
	      For k:=0 to 2 do
		  HPal[j,k]:=63-(Abs(Integer(DPal[j,k])-Integer(SPal[j,k]))-1);
	      For i:=0 to 63 do begin
		  For j:=Start to Last do
		      For k:=0 to 2 do
			  if (i>=(HPal[j,k])) then
			     Compare(SPal[j,k],DPal[j,k]);
			  if (I MOD Step)=0 then
			     SetPalettePortion(SPAl,Start,HowMany);
			  Delay(Time);
	      end;
	      SetPalettePortion(DPal,Start,HowMany);
     end;
     end;

Procedure BlendMedium(var BlendInfo:TBlendInfo);
Var i,j,k:byte;
    Last:word;
    HPal:TVGAPalette;
begin
     With BlendInfo do begin
	  Last:=Start+HowMany-1;
	  For j:=Start to Last do
	      For k:=0 to 2 do
		  HPal[j,k]:=Abs(Integer(DPal[j,k])-Integer(SPal[j,k]));
	  For i:=63 downto 0 do begin
	      For j:=Start to Last do
		  For k:=0 to 2 do
		      if (i<=(63-(63-HPal[j,k]) DIV 2)) AND
			 (i>=(63-HPal[j,k]) DIV 2) then
			 Compare(SPal[j,k],DPal[j,k]);
		      if ((63-i) MOD Step)=0 then
			 SetPalettePortion(SPal,Start,HowMany);
		      Delay(Time);
	  end;
	  SetPaletteportion(DPal,Start,HowMany);
     end;
     end;

Procedure BlendFull(var BlendInfo:TBlendInfo);
Var i,j,k:byte;
    Last:word;
    HLongPal:TLongVGAPalette;
    SLongPal:tLongVGAPalette;
begin
     With BlendInfo do begin
	  Last:=Start+HowMany-1;
	  For j:=Start to Last do
	      For k:=0 to 2 do begin
		  SLongPal[j,k]:=SPal[j,k] SHL 8;
		  HLongPal[j,k]:=((DPal[j,k]-SPal[j,k]) SHL 8) DIV 65;
		  if SPal[j,k]>DPal[j,k] then begin
		     Inc(SLongPal[j,k],255);
		     Dec(HLongPal[j,k],4);
		  end else
		      if SPal[j,k]<DPal[j,k] then Inc(HLongPal[j,k],4);
	      end;
	  For i:=0 to 63 do begin
	      For j:=Start to Last do
		  For k:=0 to 2 do begin
		      Inc(SLongPal[j,k],HLongPal[j,k]);
		      SPal[j,k]:=SLongPal[j,k] SHR 8;
		  end;
	      if (i MOD Step)=0 then
		 SetPalettePortion(SPal,Start,HowMany);
	      Delay(Time);
	  end;
	  SetPalettePortion(DPal,Start,HowMany);
     end;
     end;

Procedure MakeEffect(var Pal:TVGAPalette;EffectInfo:TEffectinfo);
Var i,j,k,l:byte;
    BlendInfo:TBlendInfo;
    BlendMethod:byte;
    HPal:TVGAPalette;
begin
     With BlendInfo do begin
	  Start:=EffectInfo.Start;
	  HowMany:=Effectinfo.HowMany;
	  Time:=EffectInfo.Time;
	  Step:=EffectInfo.Step;
	  Case EffectInfo.Effect of
	       efDark:begin
			   SPal:=Pal;
			   DPal:=BlackPal;
			   BlendMethod:=bmDown;
			   end;
	       efLight:begin
			    SPal:=BlackPal;
			    DPal:=Pal;
			    BlendMethod:=bmUp;
			    end;
	       efGrey,efColor:begin
				   For i:=Start to Start+HowMany-1 do begin
				       J:=Round(0.3*Pal[i,0]+0.59*Pal[i,1]+0.11*Pal[i,2]);
				       SetCol(HPal[i],j,j,j);
				   end;
				   if EffectInfo.Effect=efGrey then begin
				      SPal:=Pal;
				      DPal:=HPal;
				   end else begin
				       SPal:=HPal;
				       DPal:=Pal;
				   end;
				   BlendMethod:=bmFull;
				   end;
	       efBlue:begin
			   For i:=Start to HowMany-1 do begin
			       j:=Round(0.0499*Pal[i,0]);
			       k:=Round(0.19*Pal[i,1]);
			       l:=Round(1.56*Pal[i,2]);
			       SetCol(HPal[i],j,k,l);
			   end;
			   SPal:=Pal;
			   DPal:=HPal;
			   BlendMethod:=bmFull;
			   end;
	       efGreen:begin
			   For i:=Start to HowMany-1 do begin
			       j:=Round(0.05*Pal[i,0]);
			       k:=Round(1.09*Pal[i,1]);
			       l:=Round(0.0599*Pal[i,2]);
			       SetCol(HPal[i],j,k,l);
			   end;
			   SPal:=Pal;
			   DPal:=HPal;
			   BlendMethod:=bmFull;
			   end;
	       efRed:begin
			   For i:=Start to HowMany-1 do begin
			       j:=Round(1.15*Pal[i,0]);
			       k:=Round(0.19*Pal[i,1]);
			       l:=Round(0.16*Pal[i,2]);
			       SetCol(HPal[i],j,k,l);
			   end;
			   SPal:=Pal;
			   DPal:=HPal;
			   BlendMethod:=bmFull;
			   end;
	       efYellow:begin
			   For i:=Start to HowMany-1 do begin
			       j:=Round(0.35*Pal[i,0]+0.59*Pal[i,1]+0.11*Pal[i,2]);
			       SetCol(HPal[i],40,50,j);
			   end;
			   SPal:=Pal;
			   DPal:=HPal;
			   BlendMethod:=bmFull;
			   end;
	  end;
     end;
     With EffectInfo do begin
	  if Show then Blend(BlendInfo,BlendMethod);
	  if Save then Pal:=BlendInfo.DPal;
     end;
     end;

Var i:byte;

begin
For i:=0 to 255 do
    SetCol(BlackPal[i],0,0,0);
end.
